@@ -436,7 +436,11 @@ class Bus(EventEmitter):
         self.process_packet(pkt)  # handle loop-back packet
 
     def clear_attach_cache(self):
-        pass
+        self.debug("clear attach cache")
+        for d in self.devices:
+            # add a dummy byte at the end (if not done already), to force re-attach of services
+            if (len(d.services) & 3) == 0:
+                d.services.append(0)
 
     def mk_event_cmd(self, ev_code: int):
         if not self._event_counter:
@@ -1188,13 +1192,13 @@ class ServerBindings:
                 missing.append(b)
 
         sbuf = dev.services
-        for idx in range(4, len(sbuf), 4):
-            service_index = idx >> 2
+        n = dev.num_service_classes
+        for service_index in range(1, n):
             # if service is already bound to some client, move on
             if service_index in devwrap.bindings:
                 continue
 
-            service_class = util.u32(sbuf, idx)
+            service_class = dev.service_class_at(service_index)
             for i in range(len(missing)):
                 if missing[i].service_class == service_class:
                     # we've got a match!
@@ -1219,6 +1223,7 @@ class ServerBindings:
 class RoleManagerServer(Server):
     """A server for the role manager service
     """
+
     def __init__(self, bus: Bus) -> None:
         super().__init__(bus, JD_SERVICE_CLASS_ROLE_MANAGER)
 
@@ -1266,8 +1271,11 @@ class RoleManagerServer(Server):
 
     def handle_list_stored_roles(self, pkt: JDPacket):
         pipe = OutPipe(self.bus, pkt)
-        for role, payload in self.settings.list():
-            pipe.write(payload)
+        for key, payload in self.settings.list():
+            device_id, service_index = key.split(":")
+            role = jdunpack(payload, "s")[0]
+            pipe.write(bytearray(
+                jdpack("b[8] u8 s", bytearray.fromhex(device_id), service_index, role)))
         pipe.close()
 
     def handle_get_role(self, pkt: JDPacket):
@@ -1305,16 +1313,19 @@ class RoleManagerServer(Server):
             self.bus.clear_attach_cache()
             self.bind_roles()
 
-    def set_role(self, role: str, device: Optional['Device'], service_idx: Optional[int]):
-        if not device or service_idx is None:
-            self.settings.delete(role)
+    def set_role(self, role: str, device: 'Device', service_idx: int):
+        key = "{}:{}".format(device.device_id, service_idx)
+        if device:
+            self.settings.write(key, bytearray(jdpack("s", role)))
         else:
-            self.settings.write(role, bytearray(
-                jdpack("b[8] u8 s", bytearray.fromhex(device.device_id), service_idx, role)))
+            self.settings.delete(key)
+        self.bus.clear_attach_cache()
 
     def is_match_role(self, role: str, device: 'Device', service_idx: int):
-        current = self.settings.read(role)
-        return current == jdpack("b[8] u8 s", bytearray.fromhex(device.device_id), service_idx, role)
+        key = "{}:{}".format(device.device_id, service_idx)
+        current = self.settings.read(key)
+        stored_role = jdunpack(current, "s")[0] if current else None
+        return role == stored_role
 
     def _binding_hash(self):
         r = ""
@@ -1352,6 +1363,7 @@ class RoleManagerServer(Server):
                             break
                 bindings.append(b)
         servers: List[ServerBindings] = []
+
         # Group all clients by host
         for b in bindings:
             hn = b.host()
@@ -1367,6 +1379,7 @@ class RoleManagerServer(Server):
 
         # exclude hosts that have already everything bound
         servers = list(filter(lambda h: not h.fully_bound, servers))
+        self.debug("servers not fully bound: {}", len(servers))
 
         while len(servers) > 0:
             # Get host with maximum number of clients (resolve ties by name)
@@ -1375,7 +1388,8 @@ class RoleManagerServer(Server):
             for i in range(1, len(servers)):
                 a = h
                 b = servers[i]
-                if len(a.bindings) - len(b.bindings) < 0 or b.host < a.host:
+                clen = len(a.bindings) - len(b.bindings)
+                if clen < 0 or (clen == 0 and b.host < a.host):
                     h = b
             for d in wraps:
                 d.score = h.score_for(d)
@@ -1384,11 +1398,18 @@ class RoleManagerServer(Server):
             for i in range(1, len(wraps)):
                 a = dev
                 b = wraps[i]
-                if a.score - b.score < 0 or b.device.device_id < a.device.device_id:
+                cscore = a.score - b.score
+                if cscore < 0 or (cscore == 0 and b.device.device_id < a.device.device_id):
                     dev = b
 
+            self.debug("binding: server {}, device {}, score {}",
+                       h.host, dev.device.short_id, dev.score)
+            self.debug("  score: {}", ", ".join(
+                list(map(lambda w: "{}: {}".format(w.device.short_id, w.score), wraps))))
+
             if dev.score == 0:
                 # nothing can be assigned, on any device
+                self.debug("  server not bound")
                 servers.remove(h)
                 continue
 
@@ -1401,12 +1422,14 @@ class RoleManagerServer(Server):
 
             # if everything bound on this host, remove it from further consideration
             if h.fully_bound:
+                self.debug("  server bound")
                 servers.remove(h)
             else:
                 # otherwise, remove bindings on the current device, to update sort order
                 # it's unclear we need this
                 h.bindings = list(
                     filter(lambda b: b.bound_to_dev != dev.device, h.bindings))
+                self.debug("  server {} bindings", len(h.bindings))
 
         # trigger event as needed
         self._check_changes()
@@ -1712,7 +1735,7 @@ class Device(EventEmitter):
 
         role_manager = self.bus.role_manager
         if not role_manager:
-            return True
+            return False
 
         return role_manager.is_match_role(role, self, service_idx)
 
